/*
	RCIM RedHawk BenchMark test program version 2.0 for 4.x
*/
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sched.h>
#include <sys/types.h>
#include <unistd.h>
#include <mpadvise.h>
#include <cpuset.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <math.h>
#include <rcim.h>

#define MAXJITTER (30)



extern int errno;
static int BOOT_TEST=0;

void
readtime(struct timespec *nowtime)
{
        clock_gettime(CLOCK_REALTIME,nowtime);
}

void
adjust(struct timespec *start,struct timespec *finish, float *realtime)
{
        int     sec,nsec;

        sec = finish->tv_sec - start->tv_sec;
        nsec = finish->tv_nsec - start->tv_nsec;
        if (nsec<0)
        {
                sec --;
                nsec += 1000000000;
        }
        *realtime = (float)(nsec/1000)+(float)(sec*1000000);
}


#ifdef HANDLER
int loop_flag =1;
void interval_hadler(int sig)
{
        printf("interval_hadler(%d)\n",sig);
}
#endif

#define TESTSEC 10
#define MAXCOUNT 1000
void timer_test()
{
        int signum,status;
        static struct sigaction newact;
        static sigset_t set,oset;
	static siginfo_t info;
        static struct timespec t0[MAXCOUNT];
	static float real[MAXCOUNT],min,max,avr;
	static int count = 0;
	int fd,usec;
 	struct rtc rtcs;

printf("******************************************* \n");
printf("****  posix timer test wait for %d sec **** \n",TESTSEC);
printf("******************************************* \n");

	fd = open("/dev/rcim/rtc0",O_RDWR);
	if (fd<0)
	{
		fprintf(stderr,"can not open\n");
	}
	rtcs.r_modes = RTC_REPEAT;
	rtcs.r_res=USEC;
	usec = 1000*TESTSEC;
	if(ioctl(fd,RTCIOCSET,&rtcs)<0)
	{
		fprintf(stderr,"can not ioctl %s\n", strerror(errno));
		return;
	}
	if(ioctl(fd,RTCIOCSETCNT,&usec)<0)
	{
		fprintf(stderr,"can not ioctl %s\n", strerror(errno));
		return;
	}

        signum = SIGRTMIN;
	
        if (sigprocmask(0,NULL,&set)==(-1))
        {
                fprintf(stderr, "Cannot get sigprocmask - %s\n",strerror(errno));
                exit(1);
        }
        sigaddset(&set,signum);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigprocmask - %s\n", strerror(errno));
                exit(1);
        }


#ifdef HANDLER
        sigemptyset(&newact.sa_mask);
        sigaddset(&newact.sa_mask,signum);
        newact.sa_handler = interval_hadler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(signum, &newact, 0)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigaction - %s\n", strerror(errno));
                exit(1);
        }
#endif
	ioctl(fd,IOCTLSIGATTACH,signum);
	ioctl(fd,RTCIOCSTART,&status);
        do
        {
		
		ioctl(fd,RTCIOCWAIT,&status);
        	readtime(&t0[count]);
		count++;
        } while (count<MAXCOUNT);
	ioctl(fd,RTCIOCSTOP,&status);
	close(fd);
	min=1000000000;
	avr =max=0;
        for(count=1;count<MAXCOUNT;count++)
	{
                adjust(&t0[count-1],&t0[count],&real[count]);
		avr += real[count];
		if (min> real[count]) min =  real[count];
		if (max< real[count]) max =  real[count];
	}
	avr /= (float)(MAXCOUNT-1);
        printf(" min %f avr %f max %f\t",min,avr,max);
        if (abs(max-min)<=MAXJITTER)
        {
                printf("GOOD(abs(%f)<%d micro sec)\n",max-min,MAXJITTER);
        }
        else
        {
                printf("FAIL(abs(%)f>%d micro sec)\n",max-min,MAXJITTER);
        }
}

void counter_test()
{
        struct timespec t0,t1;
        float real,over;
        int i,usec,val,res,ovr,old;
	int fd,usecms,status;
 	struct rtc rtcs;


printf("*********************************************** \n");
printf("****  RCIM counter test wait              **** \n");
printf("*********************************************** \n");

	fd = open("/dev/rcim/rtc2",O_RDWR);
	if (fd<0)
	{
		fprintf(stderr,"can not open\n");
	}
	rtcs.r_modes = RTC_DEFAULT;
	rtcs.r_res=USEC; res = 1000;
	rtcs.r_clkcnt=1;
	if(ioctl(fd,RTCIOCSET,&rtcs)<0)
	{
		fprintf(stderr,"can not ioctl %s\n", strerror(errno));
		return;
	}


        printf("clock resolution is %d nsec\n",res);
	ovr = 0;
	for(i=0;i<100;i++)
	{
        	usec= 1000;
		ioctl(fd,RTCIOCSTOP,&status);
		ioctl(fd,RTCIOCSETCNT,&usec);
        	readtime(&t0);
		ioctl(fd,RTCIOCSTART,&status);
		ioctl(fd,RTCIOCWAIT,&status);
        	readtime(&t1);
        	adjust(&t0,&t1,&over);
		ovr += (((int)over) - usec);
	}
	ovr /= 100;
        printf("RTCIOCWAIT resolution test overhead is %d usec\n",ovr);
	fflush(stdout);
        for(usec=10;usec<50000;usec=usec<<1)
        {
		ioctl(fd,RTCIOCSTOP,&status);
		val = usec-ovr ;
		if (val<1) val =1;
		ioctl(fd,RTCIOCSETCNT,&val);
                readtime(&t0);
		ioctl(fd,RTCIOCSTART,&status);
		ioctl(fd,RTCIOCWAIT,&status);
                readtime(&t1);
                adjust(&t0,&t1,&real);
 		printf("request %10d usec real %10.3f usec\t",
                        usec,real);

                if (fabs(real-usec)<=MAXJITTER)
                {
                        printf("GOOD(abs(%f)<%d micro sec)\n",real-usec,MAXJITTER);
                }
                else
                {
                        printf("FAIL(abs(%f)>%d micro sec)\n",real-usec,MAXJITTER);
                }

        }
	close(fd);
}


main(int argc, char **argv)
{
	static struct sched_param myparam;
	cpuset_t* setp;
	int i,ret;
        extern int errno;
        extern int optind;
        extern char *optarg;
        register int c;
	char    *progname;


        progname = argv[0];
        while((c = getopt(argc, argv, "B")) != EOF){
                switch(c) {
                        case 'B':
				BOOT_TEST=1;
                                break;
                        default:
                                fprintf(stderr,
                                "usage: %s [-B] B:BOOT_CPU_TEST\n",
                                progname);
                                exit(1);
                }
        }
#if 0
        if (optind != argc) {
                for (c = optind; c < argc; c++) {
                        if (access(argv[c], 4)) {
                                fprintf(stderr,
                                "%s: \"%s\" %s\n",
                                 progname,argv[c],strerror(errno));
                                exit(1);
                        }
                }
        }
#endif

	mlockall(MCL_CURRENT|MCL_FUTURE);
	setp = cpuset_alloc();
        myparam.sched_priority = sched_get_priority_max(SCHED_FIFO);
        sched_setscheduler(getpid(),SCHED_FIFO,&myparam);

	system("/usr/bin/shield -a 1");
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	system("./set_affinity 2");
	cpuset_set_string(setp, "2");
	ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
	if (ret!=MPA_FAILURE)
        {
		printf("running 2nd CPU\n");
        }
	else
	{
		printf("Can not running 2nd CPU\n");
	}
	timer_test();
	counter_test();

	if (BOOT_TEST)
	{
		system("./set_affinity 1");
		cpuset_set_string(setp, "1");
		ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
		if (ret!=MPA_FAILURE)
        	{
			printf("running boot CPU\n");
        	}
		else
		{
			printf("Can not running boot CPU\n");
		}
		timer_test();
		counter_test();
	}
	system("/usr/bin/shield -r");
}
