#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>


#include <sched.h>
#include <fcntl.h>
#include <sys/vfs.h>
/* for redhawk */
#include <cpuset.h>
#include <mpadvise.h>
#include <sys/mman.h>


#define	SHARED_KEY 0x12345678


typedef struct
{
	int i,j,k;
	float a,b,c;
} shared_memory;


extern int errno;
int Global_shmid=0;
volatile sig_atomic_t Stop_flag = 0;
#ifdef SHM
shared_memory       global_buffer__=NULL;
#else
shared_memory       global_buffer__;
#endif

int global_setup_signal(int signo, void (*interrupt_handler)(int,siginfo_t *, void *))
{
        static struct sigaction newact;
        static sigset_t set,oset;

        if (sigprocmask(0,NULL,&set)==(-1))
        {
                return(-1);
        }
        sigdelset(&set,signo);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                return(-2);
        }

        sigemptyset(&newact.sa_mask);
        sigaddset(&newact.sa_mask,signo);
        newact.sa_sigaction = interrupt_handler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(signo, &newact, 0)==(-1))
        {
                return(-3);
        }
        return(0);
}

char * msg_time_get(void)
{
        time_t now_clock;
        struct tm tmp;
        static char buff[32];

        time(&now_clock);
        localtime_r((const time_t *)&now_clock,&tmp);
        asctime_r((const struct tm *)&tmp,buff);
        buff[strlen(buff)-1]=(char)':';
        buff[strlen(buff)]=(char)0;
        return(buff);
}


void print_error(char * msg,int eno)
{
        char buff[256];
        fprintf(stderr,"%s:%s - %s\n",msg_time_get(),msg,strerror_r(eno,buff,255));
}

int get_shared_memory()
{
    key_t   key;
    int     shmid,opperm_flags,shmsize,ret;
    struct shmid_ds shm_buf;

    ret = 0;
    key = SHARED_KEY;
    opperm_flags  = IPC_CREAT|SHM_R|SHM_W;
    shmsize =sizeof(shared_memory);
    shmid = shmget(key, shmsize, opperm_flags);
    if (shmid == -1)
    {
            print_error("shmget",errno);
            return(-1);
    }
    //global_buffer__ = (shared_memory *)shmat(shmid,global_buffer__,SHM_RND|SHM_REMAP);
    if ( (shared_memory *)shmat(shmid,&global_buffer__,SHM_REMAP)==(shared_memory *)NULL)
    {
            print_error("shmat",errno);
            shmctl(shmid,IPC_RMID,NULL);
            return(-1);
    }
    Global_shmid = shmid;

    if (shmctl(shmid,IPC_STAT,&shm_buf)<0)
    {
            print_error("IPC_STAT",errno);
    }
    else
    {
        shm_buf.shm_perm.mode = 0666;
        if (shmctl(shmid,IPC_SET,&shm_buf)<0)
        {
                print_error("IPC_SET",errno);
        }
    }
    if (shmctl(shmid,SHM_LOCK,NULL)<0)
    {
        fprintf(stderr,"WARNING:\t共有メモリをロックできませんでした\n");
    }
    return(ret);
}
void unget_shared_memory(void)
{
    {
#if 0
        if(shmdt((void*)&global_buffer__) == -1)
        {
            print_error("shmdt",errno);
        }
#endif
        if(shmctl(Global_shmid,IPC_RMID,NULL) == -1)
        {
            print_error("IPC_RMID",errno);
        }
    }
}

void golobal_set_envionment( int policy,int priority, int cpu_affinity_mask)
{
        struct sched_param myparam;
        cpuset_t  *cpumask;
        int cpu;
        char * cpustr;

        mlockall(MCL_CURRENT|MCL_FUTURE);

        myparam.sched_priority = priority;
        sched_setscheduler(getpid(),policy,&myparam);

        cpumask = cpuset_alloc();
        cpuset_init(cpumask);/* clear */
        for (cpu=0;cpu < 4;cpu++)
        {
                if(cpu_affinity_mask & (1<<cpu))
                {
                        cpuset_set_cpu(cpumask,cpu,1);
                }
        }
        if(mpadvise (MPA_PRC_SETBIAS,MPA_PID,getpid(),cpumask)==MPA_FAILURE)
        {
                fprintf(stderr, "%s:WARNING:\tCPUを設定できません\n" ,msg_time_get());
        }

        {
                cpuset_init(cpumask);/* clear */
                if (mpadvise (MPA_PRC_GETRUN,MPA_PID,getpid(),cpumask)!=MPA_FAILURE)
                {
                        cpustr = cpuset_get_string(cpumask);
                        printf("DEBUG:Process %d RUNNNING CPU %s\n",getpid(),cpustr);
                        free(cpustr);
                }
        }
        free(cpumask);
}


static void sigint_handler(int signo, siginfo_t *siginfo, void *misc)
{
    if ((siginfo->si_pid==0)||(getpid()==siginfo->si_pid)|| (getppid()==siginfo->si_pid))
    {
	fprintf(stderr,"\nSIGINT DETECT\n");
        Stop_flag = 1;
    }
}


main()
{
	printf("Start C program\n");
	golobal_set_envionment(SCHED_FIFO,sched_get_priority_max(SCHED_FIFO),0x8);
	global_setup_signal(SIGINT,sigint_handler);
	get_shared_memory();
	printf("Assigned from C program\nglobal_buffer__.i =0;\nglobal_buffer__.j =1;\nglobal_buffer__.k = -1;\nglobal_buffer__.a =0.0;\nglobal_buffer__.b =1.0;\nglobal_buffer__.c = -1.0;\n");
	global_buffer__.i =0;
	global_buffer__.j =1;
	global_buffer__.k = -1;
	global_buffer__.a =0.0;
	global_buffer__.b =1.0;
	global_buffer__.c = -1.0;
	printf("Call Fortran function\n");
	frt_();
	printf("Returned Fortran function\n");
	system("ipcs -m");
	printf("Please type Control-C\n");
	do
	{
		pause();
	} while(Stop_flag == 0);
	printf("End C program\n");
	unget_shared_memory();
	system("ipcs -m");
}
