/*
 * ****************************************************************************
 *  Copyright(c) 2000 Concurrent Computer Corporation
 *                     ALL RIGHTS RESERVED
 *
 *  This software is furnished under a license and may be used and copied only
 *  in accordance with the terms of such license and with the inclusion of the
 *  above copyright notice. This software or any other copies thereof, may not
 *  be provided or otherwise made available to anyone other than the licensee.
 *
 *              Title to and ownership of this software remains with
 *              the Concurrent Computer Corporation (CONCURRENT).
 *
 *       The information in this document is subject to change without
 *       notice and should not be construed as a commitment by CONCURRENT.
 *
 *       CONCURRENT assumes no responsibility for the use or reliability of
 *       its software on equipment that is not supplied by CONCURRENT.
 * ****************************************************************************
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>


#define SERVER	1
#define CLIENT	2
#define LOOP	10

volatile sig_atomic_t loop_flag =1;


void display_siginfo(siginfo_t *siginfo)
{
	printf("\tsi_signo\t%d\n\tsi_code\t\t%d\n\tsi_pid\t\t%d\n\tsi_status\t%d\n\tsi_uid\t\t%d\n\tsi_value\t%d\n\tsi_addr\t\t%d\n\tsi_fd\t\t%d\n\tsi_band\t\t%d\n\n",
	siginfo->si_signo,
	siginfo->si_code,
	siginfo->si_pid,
	siginfo->si_status,
	siginfo->si_uid,
	siginfo->si_value,
	siginfo->si_addr,
	siginfo->si_fd,
	siginfo->si_band);
	fflush(stdout);
}

void parent_intr_handler(int signo,siginfo_t *siginfo, void *val)
{
	loop_flag ++;
	printf("\nParent signal handler signo %d val %d\n",signo,*(int*)val);
	fflush(stdout);
	display_siginfo(siginfo);
}

void server_intr_handler(int signo,siginfo_t *siginfo, void *val)
{
	printf("\nServer signal handler signo %d val %d\n",signo,*(int*)val);
	fflush(stdout);
	display_siginfo(siginfo);
}

void client_intr_handler(int signo,siginfo_t *siginfo, void *val)
{
	printf("\nClient signal handler signo %d val %d\n",signo,*(int*)val);
	fflush(stdout);
	display_siginfo(siginfo);
}

void sigint_handler(int signo,siginfo_t *siginfo, void *val)
{
	printf("\nProcess %d signal handler signo %d val 0x%X\n",getpid(),signo,(long)val);
	fflush(stdout);
	display_siginfo(siginfo);
        loop_flag = 0;
}



static int
server(void)
{
	static struct sigaction newact;
	static struct sigaction intact;
	static union sigval value;
	static sigset_t imask,omask;
	static struct timespec w1 = {1,0}; /* 1 sec */
     	static siginfo_t siginfo;
	static int i,err,clientpid=-1;


	sigaddset(&newact.sa_mask,SIGRTMAX);
        newact.sa_sigaction = server_intr_handler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGRTMAX, &newact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                exit(1);
        }
	sigaddset(&intact.sa_mask,SIGINT);
        intact.sa_sigaction = sigint_handler;
        intact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGINT, &intact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                exit(1);
        }
	sigemptyset(&imask);
	sigaddset(&imask,SIGINT);
	sigaddset(&imask,SIGRTMAX);
        if (sigprocmask(SIG_BLOCK,&imask,&omask)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",strerror(errno));
                exit(1);
        }
	nanosleep(&w1,NULL);
	value.sival_int = getpid();
	printf("sigqueue from Server Process %d to Parent Process %d value %d\n",
	getpid(),getppid(), value.sival_int);
	fflush(stdout);
	sigqueue(getppid(),SIGRTMAX,value);
	while(loop_flag==1)
	{
		err = sigwaitinfo(&imask,&siginfo);
		printf("\nServer sigwaitinfo return code %d errno %d \"%s\"\n",err,errno,strerror(errno));
		display_siginfo(&siginfo);
		switch (siginfo.si_signo)
		{
			case SIGINT: loop_flag=0;break;
			case __SIGRTMAX:
				if (clientpid==(-1)&&(siginfo.si_pid==getppid())) 
				{
					clientpid = siginfo.si_int;
					loop_flag = 2;
				}
			break;
		}
	}
	sigfillset(&imask);
	sigdelset(&imask,SIGINT);
	sigdelset(&imask,SIGRTMAX);
	for(i=0;i<LOOP;i++)
	{
                sigsuspend(&imask);
	}
	for(i=0;i<LOOP;i++)
	{
		value.sival_int = i;
		sigqueue(clientpid,SIGRTMAX,value);
	}
	printf("ServerClient Process exit\n");
	fflush(stdout);
	return(0);
}


static int
client(void)
{
	static struct sigaction newact;
	static struct sigaction intact;
	static sigset_t imask,omask;
	static union sigval value;
	static struct timespec w1 = {1,0}; /* 1 sec */
     	static siginfo_t siginfo;
	static int i,err,serverpid=-1;

	sigaddset(&newact.sa_mask,SIGRTMAX);
        newact.sa_sigaction = client_intr_handler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGRTMAX, &newact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                exit(1);
        }
	sigaddset(&intact.sa_mask,SIGINT);
        intact.sa_sigaction = sigint_handler;
        intact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGINT, &intact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                exit(1);
        }
	sigemptyset(&imask);
	sigaddset(&imask,SIGINT);
	sigaddset(&imask,SIGRTMAX);
        if (sigprocmask(SIG_BLOCK,&imask,&omask)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",strerror(errno));
                exit(1);
        }
	nanosleep(&w1,NULL);
	value.sival_int = getpid();
	printf("sigqueue from Client Process %d to Parent Process %d value %d\n",
	getpid(),getppid(), value.sival_int);
	fflush(stdout);
	sigqueue(getppid(),SIGRTMAX,value);

	while(loop_flag==1)
	{
		err = sigwaitinfo(&imask,&siginfo);
		printf("\nClient sigwaitinfo return code %d errno %d \"%s\"\n",err,errno,strerror(errno));
		display_siginfo(&siginfo);
		switch (siginfo.si_signo)
		{
			case SIGINT: loop_flag=0;break;
			case __SIGRTMAX:
				if (serverpid==(-1)&&(siginfo.si_pid==getppid())) 
				{
					serverpid = (int)siginfo.si_int;
					loop_flag = 2;
				}
			break;
		}
	}
	sigfillset(&imask);
	sigdelset(&imask,SIGINT);
	sigdelset(&imask,SIGRTMAX);
	for(i=0;i<LOOP;i++)
	{
		printf("send from client data %d\n",i);
		printf("\n*******************\nany input is start");fflush(stdout); getchar();
		value.sival_int = i;
		sigqueue(serverpid,SIGRTMAX,value);
	}
	for(i=0;i<LOOP;i++)
	{
                sigsuspend(&imask);
	}
	printf("Client Process exit\n");
	fflush(stdout);
	return(0);
}

int
do_main(int option)
{
	static struct sigaction newact;
	static struct sigaction intact;
	pid_t ServerPID; 
	pid_t ClientPID; 
	int status;
	static union sigval value;
	static sigset_t imask,omask;
	static struct timespec w1 = {1,0}; /* 1 sec */

	sigaddset(&newact.sa_mask,SIGRTMAX);
        newact.sa_sigaction = parent_intr_handler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGRTMAX, &newact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                return(1);
        }

	printf("Parent Process ID %d\n",getpid());fflush(stdout);

	sigaddset(&intact.sa_mask,SIGINT);
        intact.sa_sigaction = sigint_handler;
        intact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(SIGINT, &intact, 0)==(-1))
        {
                fprintf(stderr, "Cannot set sigaction - %s\n",strerror(errno));
                return(1);
        }
	sigemptyset(&imask);
	sigaddset(&imask,SIGINT);
#ifndef TEST1
        if (sigprocmask(SIG_BLOCK,&imask,&omask)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",strerror(errno));
                return(1);
        }
#endif
	ServerPID=fork();
	if (ServerPID<0)
	{
		fprintf(stderr,"server fork error\n");
		return(0);
	}
	if (ServerPID==0)
	{
		server();
		exit(0);
	}
	printf("Server Process ID %d\n",ServerPID);fflush(stdout);
	ClientPID=fork();
	if (ClientPID<0)
	{
		fprintf(stderr,"client fork error\n");
		return(0);
	}
	if (ClientPID==0)
	{
		client();
		exit(0);
	}
	printf("Client Process ID %d\n",ClientPID);fflush(stdout);


	while(loop_flag!=3)
	{
                sigsuspend(&imask);
	}
	printf("sigqueue from Parent Process %d to Server Process %d value %d\n",getpid(),ServerPID,ClientPID);
	fflush(stdout);
	printf("\n*******************\nany input is start");fflush(stdout); getchar();
	value.sival_int = ClientPID;
	sigqueue(ServerPID,SIGRTMAX,value);
	nanosleep(&w1,NULL);



	printf("sigqueue from Parent Process %d to Client Process %d value %d\n",getpid(),ClientPID,ServerPID);
	fflush(stdout);
	value.sival_int = ServerPID;
	sigqueue(ClientPID,SIGRTMAX,value);

	waitpid(ClientPID,&status,0);
	waitpid(ServerPID,&status,0);
	printf("Parent Process exit!\n");
	return(0);
}

main( int argc, char **argv )
{
        extern int optind;
        extern char *optarg;
        register int c, option = 0;

        while((c = getopt(argc, argv, "sc")) != EOF)
        {
                switch(c)
                {
                        case 'c':
                                option|=CLIENT;
                                break;
                        case 's':
                                option|=SERVER;
                                break;
                        default:
                                break;
                }
        }
	return(do_main(option));
}
