#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <sched.h>
#include <time.h>
#include <unistd.h>
#include <sys/io.h>
#include <curses.h>
#include <poll.h>
#include <errno.h>
#include <rcim.h>


static volatile int	CYCLE = 1000;
static volatile int	loop_stop_flg = 0;
static volatile unsigned long	t_result[15] = {0,};

static struct timespec	w1us = {0, 1 * 1000};
static struct timespec	w1ms = {0, 1 * 1000 * 1000};
static struct timespec	ref_time;
static int min=0x7FFFFFFF,max=0x00000000;
static struct timespec MinCycle={0,0};
static struct timespec MaxCycle={0,0};
static unsigned int diff_max=0,diff_min=0;
static unsigned int diff_max_MAX=0;
static unsigned int diff_min_MAX=0;
static time_t ST,CT,ET;
static int use_rcim=0;

static int create_rcim_timer(int cycle)
{
	int fd,usec;
	struct rtc rtcs={0,};

        fd = open("/dev/rcim/rtc0",O_RDWR);
        if (fd<0)
        {
                fprintf(stderr,"can not open %s\n", strerror(errno));
                return(-1);
        }
        rtcs.r_modes = RTC_REPEAT;
        rtcs.r_clkcnt=USEC;
        rtcs.r_res=USEC;
        if(ioctl(fd,RTCIOCSET,&rtcs)<0)
        {
                fprintf(stderr,"can not ioctl %s\n", strerror(errno));
                return(-1);
        }
	usec = cycle;
        if(ioctl(fd,RTCIOCSETCNT,&usec)<0)
        {
                fprintf(stderr,"can not ioctl %s\n", strerror(errno));
                return(-1);
        }
	return(fd);
}

static timer_t create_posix_timer(int cycle)
{
	timer_t		tim;
	struct sigevent	event;
	struct itimerspec	itspec;

	memset(&event, 0, sizeof(event));
	event.sigev_notify = SIGEV_SIGNAL;
	event.sigev_signo  = SIGRTMIN;
	if (timer_create((clockid_t)CLOCK_REALTIME, &event, &tim) == (-1))
	{
		return (timer_t)(-1);
	}

	clock_gettime(CLOCK_REALTIME, &itspec.it_value);
	itspec.it_value.tv_sec  += 1;
	itspec.it_value.tv_nsec += 0;
	itspec.it_interval.tv_sec  = 0;
	itspec.it_interval.tv_nsec = 1000 * cycle;
	if (timer_settime(tim, TIMER_ABSTIME, &itspec, NULL) == (-1))
	{
		return (timer_t)(-1);
	}

	return tim;
}

static int adjust(struct timespec *start, struct timespec *end, int *min,int *max)
{
	int		sec, nsec, rtn;
	unsigned int	diff;
        static struct timespec  t1;

	sec  = end->tv_sec - start->tv_sec;
	nsec = end->tv_nsec - start->tv_nsec;
	if (nsec < 0)
	{
		sec--;
		nsec += 1000000000;
	}

	diff = (unsigned long)((sec * 1000000) + (nsec / 1000));

	if (diff < (CYCLE - 90))	
	{
        	clock_gettime(CLOCK_REALTIME, &t1);
        	sec  = t1.tv_sec - MinCycle.tv_sec;
        	nsec = t1.tv_nsec - MinCycle.tv_nsec;
		MinCycle=t1;
        	if (nsec < 0)
        	{
                	sec--;
                	nsec += 1000000000;
        	}
        	diff_min = (unsigned long)((sec * 1000000) + (nsec / 1000));
		if (diff_min_MAX<diff_min) diff_min_MAX = diff_min;
		if (diff<*min) *min=diff;
		rtn =  0;
	}
	else if (diff <  (CYCLE - 75))	rtn =  1;
	else if (diff <  (CYCLE - 60))	rtn =  2;
	else if (diff <  (CYCLE - 45))	rtn =  3;
	else if (diff <  (CYCLE - 30))	rtn =  4;
	else if (diff <  (CYCLE - 15))	rtn =  5;
	else if (diff <  CYCLE)		rtn =  6;
	else if (diff == CYCLE)		rtn =  7;
	else if (diff <= (CYCLE + 15))	rtn =  8;
	else if (diff <= (CYCLE + 30))	rtn =  9;
	else if (diff <= (CYCLE + 45))	rtn = 10;
	else if (diff <= (CYCLE + 60))	rtn = 11;
	else if (diff <= (CYCLE + 75))	rtn = 12;
	else if (diff <= (CYCLE + 90))	rtn = 13;
	else
	{
        	clock_gettime(CLOCK_REALTIME, &t1);
        	sec  = t1.tv_sec - MaxCycle.tv_sec;
        	nsec = t1.tv_nsec - MaxCycle.tv_nsec;
		MaxCycle=t1;
        	if (nsec < 0)
        	{
                	sec--;
                	nsec += 1000000000;
        	}
        	diff_max = (unsigned long)((sec * 1000000) + (nsec / 1000));
		if (diff_max_MAX<diff_max) diff_max_MAX = diff_max;
		if (diff>*max) *max=diff;
	        rtn = 14;
	}
	return rtn;
}

static void sigint_handler(int sig)
{
	loop_stop_flg = 1;
}


static void sigrtmin_handler(int sig)
{
	static unsigned long	counter = 0;
	static struct timespec	t0, t1;

	clock_gettime(CLOCK_REALTIME, &t1);
	if (counter > 0)
	{
		t_result[adjust(&t0, &t1, &min,&max)]++;
	}

	t0 = t1;
	counter++;
}

void make_disp(void)
{
	/*                    0123456789012345678901234567890123456789*/
	move( 0,  0); printw("  Frequency:   %dHz", 1000000 / CYCLE);
	move( 1,  0); printw("     Histogram         Count         Value   ShortCycle    LongCycle");
	move( 2,  0); printw(" Under  %5dus", CYCLE - 90);
	move( 3,  0); printw(" %5d-%5dus ", CYCLE - 90, CYCLE - 76);
	move( 4,  0); printw(" %5d-%5dus ", CYCLE - 75, CYCLE - 61);
	move( 5,  0); printw(" %5d-%5dus ", CYCLE - 60, CYCLE - 46);
	move( 6,  0); printw(" %5d-%5dus ", CYCLE - 45, CYCLE - 31);
	move( 7,  0); printw(" %5d-%5dus ", CYCLE - 30, CYCLE - 16);
	move( 8,  0); printw(" %5d-%5dus ", CYCLE - 15, CYCLE -  1);
	move( 9,  0); printw("       %5dus ", CYCLE);
	move(10,  0); printw(" %5d-%5dus ", CYCLE +  1, CYCLE + 15);
	move(11,  0); printw(" %5d-%5dus ", CYCLE + 16, CYCLE + 30);
	move(12,  0); printw(" %5d-%5dus ", CYCLE + 31, CYCLE + 45);
	move(13,  0); printw(" %5d-%5dus ", CYCLE + 46, CYCLE + 60);
	move(14,  0); printw(" %5d-%5dus ", CYCLE + 61, CYCLE + 75);
	move(15,  0); printw(" %5d-%5dus ", CYCLE + 76, CYCLE + 90);
	move(16,  0); printw(" Over  %5dus", CYCLE + 90);
}

void do_disp(int flg)
{
	int	n,cur;

	if (flg == 0) return;
 	time(&CT);
	cur = CT-ST;
	move(0, 30); printw("%02d:%02d:%02d", cur/3600,(cur%3600)/60,cur%60);
	for (n = 0; n < 15; n++)
	{
		move(2 + n, 17);
		printw("%10ld", t_result[n]);
	}
	if (t_result[ 0]) 
	{
		move(2 + 0, 30); printw("%10ldus %10ldus %10ldus", min , diff_min,diff_min_MAX);
	}
	if (t_result[14]) 
	{
		move(2 +14, 30); printw("%10ldus %10ldus %10ldus", max , diff_max,diff_max_MAX);
	}
}

void do_main(void)
{
	int		key, nfds = 1;
	int		disp_flg = 1;
	struct pollfd	fdarray[1];
	int cur;

	initscr();
	cbreak();
	noecho();
	nonl();
	timeout(1);
	printf("\x1b[?25l");

	fdarray[0].fd = STDIN_FILENO;
	fdarray[0].events = POLLIN | POLLPRI;

 	time(&ST);
	make_disp();
        clock_gettime(CLOCK_REALTIME, &MinCycle);
        clock_gettime(CLOCK_REALTIME, &MaxCycle);
	while (loop_stop_flg == 0)
	{
		if (poll(fdarray, nfds, 100) < 0)
		{
			nanosleep(&w1us, NULL);
		}
		else
		{
			key = getch();
			if (key == 'q' || key == 'Q')
			{
				break;
			}
			else if (key == 'p' || key == 'P')
			{
				disp_flg = (~disp_flg) & 0x01;
			}
		}
		do_disp(disp_flg);
		refresh();
	}

	printf("\x1b[?25h");
	clear();
	refresh();
	echo();
	nl();
	endwin();

        printf("  Frequency:   %dHz\n", 1000000 / CYCLE);
        printf("     Histogram         Count         Value   ShortCycle    LongCycle\n");
        printf(" Under %5dus %10ld", CYCLE - 90,t_result[0]);
 	if (t_result[0]) 
		printf(" %10ldus %10ldus %10ldus%d\n", min , diff_min,diff_min_MAX);
	else
		printf("\n");

        printf(" %5d-%5dus %10ld\n", CYCLE - 90, CYCLE - 76,t_result[1]);
        printf(" %5d-%5dus %10ld\n", CYCLE - 75, CYCLE - 61,t_result[2]);
        printf(" %5d-%5dus %10ld\n", CYCLE - 60, CYCLE - 46,t_result[3]);
        printf(" %5d-%5dus %10ld\n", CYCLE - 45, CYCLE - 31,t_result[4]);
        printf(" %5d-%5dus %10ld\n", CYCLE - 30, CYCLE - 16,t_result[5]);
        printf(" %5d-%5dus %10ld\n", CYCLE - 15, CYCLE -  1,t_result[6]);
        printf("       %5dus %10ld\n", CYCLE,t_result[7]);
        printf(" %5d-%5dus %10ld\n", CYCLE +  1, CYCLE + 15,t_result[8]);
        printf(" %5d-%5dus %10ld\n", CYCLE + 16, CYCLE + 30,t_result[9]);
        printf(" %5d-%5dus %10ld\n", CYCLE + 31, CYCLE + 45,t_result[10]);
        printf(" %5d-%5dus %10ld\n", CYCLE + 46, CYCLE + 60,t_result[11]);
        printf(" %5d-%5dus %10ld\n", CYCLE + 61, CYCLE + 75,t_result[12]);
        printf(" %5d-%5dus %10ld\n", CYCLE + 76, CYCLE + 90,t_result[13]);
        printf(" Over  %5dus %10ld", CYCLE + 90,t_result[14]);
 	if (t_result[14]) 
		printf(" %10ldus %10ldus %10ldus\n", max , diff_max,diff_max_MAX);
	else
		printf("\n");
 	time(&ET);
	cur = ET-ST;
	printf("Total Test Time  %02d:%02d:%02d\n", cur/3600,(cur%3600)/60,cur%60);

}

int main(int argc, char *argv[])
{
	int		c;
	extern char	*optarg;
	struct sigaction	sigact;
	struct sched_param	sparam;
	timer_t		pos_timer;
	int	rcim_fd;
	int status;
	static sigset_t set,oset;

	while((c = getopt(argc, argv, "hc:r")) != EOF)
	{
		if (c == 'c')
		{
			/* CYCLE: Interval[usec] */
			CYCLE = 1000000 / atoi(optarg);
		}
		else if (c == 'r')
		{
			use_rcim=1;
		}
		else if (c == 'h')
		{
			printf("usage: %s -c [Cycle(Hz)]\n\n",argv[0]);
			exit(1);
		}
	}

	mlockall(MCL_CURRENT | MCL_FUTURE);

	sched_getparam(0, &sparam);
	sparam.sched_priority = 10;
	sched_setscheduler(0, SCHED_FIFO, (const struct sched_param*)&sparam);

	sigemptyset(&sigact.sa_mask);
	sigaddset(&sigact.sa_mask, SIGINT);
	sigact.sa_flags   = SA_RESTART;
	sigact.sa_handler = sigint_handler;
	if (sigaction(SIGINT, &sigact, NULL) < 0)
	{
		exit(1);
	}

	sigemptyset(&sigact.sa_mask);
	sigaddset(&sigact.sa_mask, SIGRTMIN);
	sigact.sa_flags   = SA_RESTART;
	sigact.sa_handler = sigrtmin_handler;
	if (sigaction(SIGRTMIN, &sigact, NULL) < 0)
	{
		exit(1);
	}

	if (use_rcim)
	{
		sigemptyset(&set);
        	sigaddset(&set,SIGRTMIN);
        	if (sigprocmask(SIG_UNBLOCK,&set,&oset)==(-1))
        	{
                	fprintf(stderr, "%s: Cannot set sigprocmask - %s\n", strerror(errno));
                	exit(1);
        	}
		rcim_fd = create_rcim_timer(CYCLE);
		if (rcim_fd<0) exit(0);
        	ioctl(rcim_fd,IOCTLSIGATTACH,SIGRTMIN);
        	ioctl(rcim_fd,RTCIOCSTART,&status);
	}
	else
	{
		pos_timer = create_posix_timer(CYCLE);
	}
	if (pos_timer < 0)
	{
		exit(1);
	}

	do_main();

	if (use_rcim)
	{
		ioctl(rcim_fd,RTCIOCSTOP,&status);
		close(rcim_fd);
	}
	else
	{	
		timer_delete(pos_timer);
	}
	munlockall();

	exit(0);
}


